#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <vector>

/* TODO: fix those */
#include "camera.h"
#include "mesh.h"
#include "object.h"
#include "scene.h"

#include "opengl/opengl.h"
#include "vulkan/vk.h"

/* static functions */

static bool init();
static void cleanup();
static void display();

/* glfw callbacks */
static void key_clbk(GLFWwindow *win, int key, int scancode, int action, int mods);
static void motion_clbk(GLFWwindow *win, double x, double y);
static void mouse_clbk(GLFWwindow *win, int button, int action, int mods);

/* global variables */
bool use_vulkan;

GLFWwindow *win;
int win_w = 800;
int win_h = 600;

Camera *camera;

/* variables */
// TODO: remove just for test:
static Scene scene;

int main(int argc, char **argv)
{
	for(int i=0; i<argc; ++i) {
		if(strcmp(argv[i], "-opengl") == 0) {
			use_vulkan = false;
			printf("Backend: OpenGL.\n");
		}
		else if(strcmp(argv[i], "-vulkan") == 0) {
			use_vulkan = true;
			printf("Backend: Vulkan.\n");
		}
		else {
			printf("No backend specified. Using OpenGL.\n");
		}
	}

	if(!init()) {
		fprintf(stderr, "Failed to initialize program.\n");
		return 1;
	}

	glfwSetKeyCallback(win, key_clbk);
	glfwSetCursorPosCallback(win, motion_clbk);
	glfwSetMouseButtonCallback(win, mouse_clbk);

	while(!glfwWindowShouldClose(win)) {
		display();

		glfwSwapBuffers(win);
		glfwPollEvents();
	}

	atexit(cleanup);
	return 0;
}

static bool init()
{
	if(use_vulkan) {
		if(!init_vulkan())
			return false;
	}
	else {
		if(!init_opengl())
			return false;
	}

	camera = new Camera(25, 25, 4, 45);

	if(!scene.load("data/spot/spot_control_mesh.obj")) {
		fprintf(stderr, "Failed to load scene.\n");
		return false;
	}

	for(size_t i=0; i<scene.objects.size(); ++i) {
		printf("object: %d\n", (int)i);
		printf("mesh: %s\n", scene.objects[i]->mesh->name.c_str());
		printf("material: %s\n", scene.objects[i]->material->name.c_str());
		printf("transform:\n");
		scene.objects[i]->transform.print();
	}
	return true;
}

static void cleanup()
{
	if(use_vulkan) {
		cleanup_vulkan();
	}
	else {
		cleanup_opengl();
	}

	delete camera;
}

static void key_clbk(GLFWwindow *win, int key, int scancode, int action, int mods)
{
	if(key == GLFW_KEY_ESCAPE && action == GLFW_PRESS) {
		glfwSetWindowShouldClose(win, GLFW_TRUE);
	}
}

static double prev_x, prev_y;
static int bnstate[8];

static void motion_clbk(GLFWwindow *win, double x, double y)
{
	int dx = x - prev_x;
	int dy = y - prev_y;

	prev_x = x;
	prev_y = y;

	if(!dx && !dy) return;

	if(bnstate[0]) {
		camera->theta += dx * 0.5;
		camera->phi += dy * 0.5;

		if(camera->phi < -90) camera->phi = -90;
		if(camera->phi > 90) camera->phi = 90;
	}
	if(bnstate[2]) {
		camera->distance += dy * 0.1;
		if(camera->distance < 0.0) camera->distance = 0.0;
	}
}

static void mouse_clbk(GLFWwindow *win, int bn, int action, int mods)
{
	bnstate[bn - GLFW_MOUSE_BUTTON_LEFT] = action == GLFW_PRESS ? 1 : 0;
	glfwGetCursorPos(win, &prev_x, &prev_y);
}

static void display()
{
	if(use_vulkan) {
		display_vulkan();
	}
	else {
		display_opengl();
	}
}
